--- LibGroup
-- @class file
-- @name LibGroup-1.0.lua
-- @release 1.0
--
-- Events fired by this library:
--   Joined [name] - fired whenever anyone joins the raid (including ourself)
--   Left [name] - fired whenever anyone leaves the raid (including ourself)
--   Connected [name] - fired whenever anyone connects after a disconnect
--   Disconnected [name] - fired whenever anyone disconnects
--   LeaderChanged [newName, oldName] - fired whenever the group leader changes
-- Raid only events:
--   AssistAdded [name] - fired whenever anyone gains assist
--   AssistRemoved [name] - fired whenever anyone loses assist
--   MasterLooterChanged [newName, oldName] - fired whenever the master looter changes

--[[
   Copyright 2009 Christo Kotze (onaforeignshore)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
]]

assert(LibStub:GetLibrary("CallbackHandler-1.0", true), "LibGroup-1.0 requires CallbackHandler-1.0")

local MAJOR_VERSION = "LibGroup-1.0"
local MINOR_VERSION = tonumber(("$revision: 3 $"):match("%d+")) or 0

local lib, oldMinor = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

lib.callbacks = lib.callbacks or LibStub:GetLibrary("CallbackHandler-1.0"):New(lib)
local callbacks = lib.callbacks

-- localise global functions and variables
local GetNumRaidMembers = _G.GetNumRaidMembers
local GetRaidRosterInfo = _G.GetRaidRosterInfo
local GetNumPartyMembers = _G.GetNumPartyMembers
local UnitName = _G.UnitName
local UnitIsPartyLeader = _G.UnitIsPartyLeader
local UnitLevel = _G.UnitLevel
local UnitClass = _G.UnitClass
local UnitIsConnected = _G.UnitIsConnected
local UnitIsDeadOrGhost = _G.UnitIsDeadOrGhost

-- local variables
local new, del -- to make cache functions local
local inGroup = false -- do test if we're in a group
local group = {} -- the group table
local groupType -- the type of group
local groupLeader -- the name of the group leader
local masterLooter -- the name of the master looter (raid only)

-- create a table cache
do
	-- unused tables go in this set
	-- if the garbage collector comes around, they'll be collected properly
	local cache = setmetatable({}, {__mode='k'})
	
	--- Return a table
	-- @usage local t = new()
	-- @return a blank table
	function new()
		local t = next(cache)
		if t then
			cache[t] = nil
			return t
		end
		
		return {}
	end
	
	--- Delete a table, clearing it and putting it back into the queue
	-- @usage local t = new()
	-- t = del(t)
	-- @return nil
	function del(t)
		wipe(t)
		cache[t] = true
		return nil
	end
end

-- create a frame for the library
local frame
if lib.frame then
	frame = lib.frame
	frame:UnregisterAllEvents()
	frame:SetScript("OnEvent", nil)
	frame:SetScript("OnUpdate", nil)
else
	frame = CreateFrame("Frame", MAJOR_VERSION .. "-Frame")
	lib.frame = frame
end
frame:SetScript("OnEvent", function(this, event, ...)
	this[event](lib, ...)
end)
-- register RAID and PARTY events
frame:RegisterEvent("RAID_ROSTER_UPDATE")
frame:RegisterEvent("PARTY_MEMBERS_CHANGED")

-- called whenever the raid roster changes
-- @usage internal use only
function frame:RAID_ROSTER_UPDATE()
	if GetNumRaidMembers() >= 1 then
		for i = 1, GetNumRaidMembers() do
			local id
			id = "raid"..i
			local name, rank, subgroup, level, _, fileName, _, online, dead, _, isML = GetRaidRosterInfo(i)
			-- used for cross realm groups
			local _, server = UnitName(id)
			if server and server ~= ""  then
				name = name.."-"..server
			end
			local guid = UnitGUID(id)
			-- if a player isn't in our table, add him/her
			if (not group[name]) and inGroup then
				callbacks:Fire("Joined", name)
			end
			if dead == "" then dead = nil end
			if name then
				group[name] = group[name] or {}
				-- need to store the name, as the key may be cross realm
				group[name].name = name
				local oldRank = group[name].rank
				group[name].rank = rank
				-- if the player wasn't an assistant, but is now
				if rank == 1 and oldRank ~= 1 then
					callbacks:Fire("AssistAdded", name)
				end
				-- if the player was an assistant, but isn't anymore
				if rank == 0 and oldRank == 1 then
					callbacks:Fire("AssistRemoved", name)
				end
				if rank == 2 then
					-- if the raid leader has changed
					if oldRank ~= 2 then
						callbacks:Fire("LeaderChanged", name, groupLeader)
					end
					groupLeader = name
				end
				group[name].level = level
				group[name].subgroup = subgroup
				group[name].class = fileName
				-- if the player was online, but isn't anymore
				if group[name].online == 1 and not online then
					callbacks:Fire("Disconnected", name)
				end
				-- if the player was offline, but isn't anymore
				if group[name].online == -1 and online then
					callbacks:Fire("Connected", name)
				end
				group[name].online = online or -1 -- trickery to get above callbacks to work
				group[name].dead = dead
				if isML and name ~= masterLooter then
					-- if the master looter has changed
					callbacks:Fire("MasterLooterChanged", name, masterLooter)
					masterLooter = name
				end
				group[name].id = id
				group[name].guid = guid
				-- flag to show we've updated this player. any unflagged members are assumed to have left
				group[name].updated = true
			end
		end
		-- check to see if we have just joined the raid
		if not inGroup then
			inGroup = true
			groupType = "RAID"
			callbacks:Fire("Joined", UnitName("player"))
		end
		-- check to see who has left the raid
		for i, v in pairs(group) do
			if not v.updated then
				group[i] = nil
				callbacks:Fire("Left", i)
			else
				v.updated = nil
			end
		end
	else -- we've left the group - clean up
		inGroup = false
		groupType = nil
		groupLeader = nil
		masterLooter = nil
		callbacks:Fire("Left", UnitName("player"))
		-- clear the table for next time
		for i, _ in pairs(group) do
			group[i] = nil
		end
	end
end

-- called whenever the party members change
-- @usage internal use only
function frame:PARTY_MEMBERS_CHANGED()
	if GetNumRaidMembers() > 0 then return end
	if GetNumPartyMembers() >= 1 then
		for i = 0, GetNumPartyMembers() do
			local id
			if (i == 0) then
				id = "player"
			else
				id = "party"..i
			end
			-- used for cross realm groups
			local name, server = UnitName(id)
			if server and server ~= ""  then
				name = name.."-"..server
			end
			local rank = UnitIsPartyLeader(id)
			local level = UnitLevel(id)
			local _, fileName = UnitClass(id)
			local online = UnitIsConnected(id)
			local dead = UnitIsDeadOrGhost(id)
			if dead == "" then dead = nil end
			local guid = UnitGUID(id)
			if (not group[name]) and inGroup then
				callbacks:Fire("Joined", name)
			end
			if name then
				group[name] = group[name] or {}
				-- need to store the name, as the key may be cross realm
				group[name].name = name
				local oldRank = group[name].rank
				if rank then
					group[name].rank = 2
				else
					group[name].rank = 0
				end
				if rank == 2 then
					groupLeader = name
					if oldRank ~= 2 then
						callbacks:Fire("LeaderChanged", name)
					end
				end
				group[name].level = level
				group[name].class = fileName
				-- if the player was online, but isn't anymore
				if group[name].online == 1 and not online then
					callbacks:Fire("Disconnected", name)
				end
				-- if the player was offline, but isn't anymore
				if group[name].online == -1 and online then
					callbacks:Fire("Connected", name)
				end
				group[name].online = online or -1 -- trickery to get above callbacks to work
				group[name].dead = dead
				group[name].id = id
				group[name].guid = guid
				-- flag to show we've updated this player. any unflagged members are assumed to have left
				group[name].updated = true
			end
		end
		if not inGroup then
			inGroup = true
			groupType = "PARTY"
			callbacks:Fire("Joined", UnitName("player"))
		end
		for i, v in pairs(group) do
			if not v.updated then
				group[i] = nil
				callbacks:Fire("Left", i)
			else
				v.updated = nil
			end
		end
	else
		inGroup = false
		groupType = nil
		groupLeader = nil
		masterLooter = nil
		callbacks:Fire("Left", UnitName("player"))
		-- clear the table for next time
		for i, _ in pairs(group) do
			group[i] = nil
		end
	end
end

--- check if we are in a group
-- @returns true if we are in a group
function lib:IsInGroup()
	return inGroup
end

--- find a player's class
-- @param name - player's name (name-server for cross realm)
-- @returns class or "UNKNOWN"
function lib:GetClass(name)
	name = name or UnitName("player")
	return (group[name] and group[name].class) or "UNKNOWN"
end

--- get a player's class color (decimal)
-- @param name - player's name (name-server for cross realm)
-- @returns r, g, b
function lib:GetClassColor(name)
	name = name or UnitName("player")
	local class = group[name] and group[name].class
	if not class then return 0.8, 0.8, 0.8 end
	local c = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS
	return c.r, c.g, c.b
end

--- get a player's class color (hexadecimal)
-- @param name - player's name (name-server for cross realm)
-- @returns class color string
function lib:GetClassHexColor(name)
	name = name or UnitName("player")
	local class = group[name] and group[name].class
	if not class then return '|cffcccccc' end
	local c = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS[class]
	return ("|cff%02X%02X%02X"):format(c.r * 255, c.g * 255, c.b * 255)
end

--- find the group leader
-- @returns group leader's name
function lib:GetGroupLeader()
	return groupLeader
end

--- find the group type
-- @returns "RAID", "PARTY" or nil
function lib:GetGroupType()
	return groupType
end

--- find a player's level
-- @param name - player's name (name-server for cross realm)
-- @returns the level or 0 if offline
function lib:GetLevel(name)
	name = name or UnitName("player")
	return (group[name] and group[name].level) or 0
end

--- find a player's rank
-- @param name - player's name (name-server for cross realm)
-- @returns 2 for RL or PL, 1 for RA, or 0
function lib:GetRank(name)
	name = name or UnitName("player")
	return (group[name] and group[name].rank) or 0
end

--- find a player's subgroup (only in raid)
-- @param name - player's name (name-server for cross realm)
-- @returns subgroup or 0
function lib:GetSubgroup(name)
	name = name or UnitName("player")
	return (group[name] and group[name].subgroup) or 0
end

--- find a player's unit id
-- @param name - player's name (name-server for cross realm)
-- @returns the unit id or "none"
function lib:GetUnitId(name)
	name = name or UnitName("player")
	return (group[name] and group[name].id) or "none"
end

--- find a player's GUID (Globally Unique ID)
-- @param name - player's name (name-server for cross realm)
-- @returns the GUID string or 0
function lib:GetUnitGUID(name)
	name = name or UnitName("player")
	return (group[name] and group[name].guid) or 0
end

	
--- check if a player is dead
-- @param name - player's name (name-server for cross realm)
-- @returns true if dead
function lib:IsDead(name)
	name = name or UnitName("player")
	return (group[name] and group[name].dead) or 0
end

--- check if a player is the master looter
-- @param name - player's name (name-server for cross realm)
-- @returns true if player is the master looter
function lib:IsMasterLooter(name)
	name = name or UnitName("player")
	return name == masterLooter or 0
end

--- check if a player is online
-- @param name - player's name (name-server for cross realm)
-- @returns true if the player is online
function lib:IsOnline(name)
	name = name or UnitName("player")
	return (group[name] and group[name].online) or 0
end

--- get the player's name colored in his/her class
-- @param name - player's name (name-server for cross realm)
-- @returns the class colored player's name
function lib:GetPlayersWithClassColor()
	local t = {}
	for k, _ in pairs(group) do
		t[k] = "|cff"..lib:GetClassHexColor(k)..k.."|r"
	end
	return t
end

local sorts; sorts = {
	NAME = function(a, b)
		return group[a].name < group[b].name
	end,
	LEVEL = function(a, b)
		if group[a].level == group[b].level then
			return sorts.NAME(a, b)
		else
			return group[a].level < group[b].level
		end
	end,
	CLASS = function(a, b)
		if group[a].class == group[b].class then
			return sorts.NAME(a, b)
		else
			return group[a].class < group[b].class
		end
	end,
	RANK = function(a, b)
		if group[a].rank == group[b].rank then
			return sorts.NAME(a, b)
		else
			return group[a].rank < group[b].rank
		end
	end,
	ONLINE = function(a, b)
		if group[a].online == group[b].online then
			return sorts.NAME(a, b)
		else
			return group[a] -- 1 or 0 (nil), you see.
		end
	end,
}

local iter = function(t)
	local n = t.n + 1
	t.n = n
	return t[n] or del(t)
end

--- iterate through the players
-- @param includeOffline - true to include offline players
-- @param sort - "NAME", "LEVEL", "CLASS", "RANK", "ONLINE" - defaults to "NAME" if nil
function lib:Iterate(includeOffline, sort)
	local sortFunc = sorts[sort or "NAME"]
	if not sortFunc then
		error('Argument #2 must be "NAME", "LEVEL", "CLASS", "RANK", "ONLINE", or nil.')
	end
	
	local tmp = new()
	for _, v in pairs(group) do
		if includeOffline then
			tmp[#tmp + 1] = v.name
		else
			if v.online then tmp[#tmp + 1] = v.name end
		end
	end
	
	table.sort(tmp, sortFunc)
	tmp.n = 0
	
	return iter, tmp, nil
end

if inGroup then
	frame:Show()
else
	frame:Hide()
end